@@ -1,7 +1,7 @@
 /* This file is part of the KDE project
    Copyright (C) 2004 Cedric Pasteur <cedric.pasteur@free.fr>
    Copyright (C) 2004 Alexander Dymo <cloudtemple@mskat.net>
-   Copyright (C) 2004-2015 Jarosław Staniek <staniek@kde.org>
+   Copyright (C) 2004-2017 Jarosław Staniek <staniek@kde.org>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -20,123 +20,248 @@
 */
 
 #include "KProperty.h"
-#include "KProperty_p.h"
-#include "KPropertySet.h"
 #include "KPropertyFactory.h"
+#include "KPropertyListData.h"
+#include "KPropertySet_p.h"
+#include "KProperty_p.h"
 #include "kproperty_debug.h"
 
-#include <QPointer>
+#include <math.h>
 
-//! @internal
-class KProperty::Private
+//! @return true if @a currentValue and @a value are compatible
+static bool compatibleTypes(const QVariant& currentValue, const QVariant &value)
 {
-public:
-    Private()
-            : type(KProperty::Auto), caption(0), listData(0), changed(false), storable(true),
-            readOnly(false), visible(true),
-            autosync(-1), composed(0), useComposedProperty(true),
-            sets(0), parent(0), children(0), relatedProperties(0)
+    if (currentValue.isNull() || value.isNull())
+        return true;
+    const QVariant::Type t = currentValue.type();
+    const QVariant::Type newt = value.type();
+    if (t == newt)
+        return true;
+    if (   (t == QVariant::Int && newt == QVariant::UInt)
+        || (t == QVariant::UInt && newt == QVariant::Int)
+        || (t == QVariant::ByteArray && newt == QVariant::String)
+        || (t == QVariant::String && newt == QVariant::ByteArray)
+        || (t == QVariant::ULongLong && newt == QVariant::LongLong)
+        || (t == QVariant::LongLong && newt == QVariant::ULongLong))
     {
+        return true;
+    }
+    return false;
+}
+
+// ----
+
+KProperty::Private::Private(KProperty *prop)
+        : q(prop), type(KProperty::Auto), listData(nullptr), changed(false), storable(true),
+        readOnly(false), visible(true),
+        composed(nullptr), useComposedProperty(true),
+        sets(nullptr), parent(nullptr), children(nullptr), relatedProperties(nullptr)
+{
+}
+
+void KProperty::Private::setCaptionForDisplaying(const QString& captionForDisplaying)
+{
+    caption = captionForDisplaying.simplified();
+    if (caption == captionForDisplaying) {
+        caption.clear();
+    }
+    this->captionForDisplaying = captionForDisplaying;
+}
+
+KProperty::Private::~Private()
+{
+    delete listData;
+    if (children) {
+        qDeleteAll(*children);
+        delete children;
+    }
+    delete relatedProperties;
+    delete composed;
+    delete sets;
+}
+
+bool KProperty::Private::valueDiffersInternal(const QVariant &otherValue, KProperty::ValueOptions options)
+{
+    if (!compatibleTypes(value, otherValue)) {
+        kprWarning() << "INCOMPATIBLE TYPES! old=" << value << "new=" << otherValue << "in property" << q->name();
     }
 
-    inline void setCaptionForDisplaying(const QString& captionForDisplaying) {
-        delete caption;
-        if (captionForDisplaying.simplified() != captionForDisplaying) {
-            if (captionForDisplaying.isEmpty()) {
-                caption = 0;
-            }
-            else {
-               caption = new QString(captionForDisplaying.simplified());
+    const QVariant::Type t = value.type();
+    const QVariant::Type newt = otherValue.type();
+    if (   t == QVariant::DateTime
+        || t == QVariant::Time)
+    {
+        //for date and datetime types: compare with strings, because there
+        //can be miliseconds difference
+        return value.toString() != otherValue.toString();
+    }
+    else if (t == QVariant::String || t == QVariant::ByteArray) {
+        //property is changed for string type,
+        //if one of value is empty and other isn't..
+        return (value.toString().isEmpty() != otherValue.toString().isEmpty())
+              //..or both are not empty and values differ
+              || (!value.toString().isEmpty() && !otherValue.toString().isEmpty() && value != otherValue);
+    }
+    else if (t == QVariant::Double) {
+        const double factor = pow(10.0, option("precision", KPROPERTY_DEFAULT_DOUBLE_VALUE_PRECISION).toDouble());
+        //kprDebug()
+        //    << "factor:" << factor << "precision:" << option("precision", KPROPERTY_DEFAULT_DOUBLE_VALUE_STEP)
+        //    << "double compared:" << value.toDouble() << otherValue.toDouble()
+        //    << ":" << static_cast<qlonglong>(value.toDouble() * factor) << static_cast<qlonglong>(otherValue.toDouble() * factor);
+        return static_cast<qlonglong>(value.toDouble() * factor) != static_cast<qlonglong>(otherValue.toDouble() * factor);
+    } else if (t == QVariant::Invalid && newt == QVariant::Invalid) {
+        return false;
+    } else if (composed && !(options & ValueOption::IgnoreComposedProperty)) {
+        return !composed->valuesEqual(value, otherValue);
+    }
+    else {
+        return value != otherValue;
+    }
+}
+
+bool KProperty::Private::setValueInternal(const QVariant &newValue, KProperty::ValueOptions valueOptions)
+{
+    if (name.isEmpty()) {
+        kprWarning() << "COULD NOT SET value to a null property";
+        return false;
+    }
+
+    //1. Check if the value should be changed
+    if (!valueDiffersInternal(newValue, valueOptions)) {
+        return false;
+    }
+
+    //2. Then change it, and store old value if necessary
+    if (valueOptions & KProperty::ValueOption::IgnoreOld) {
+        oldValue = QVariant(); // clear old value
+        changed = false;
+    } else {
+        if (!changed) {
+            oldValue = value;
+            changed = true;
+        }
+    }
+    if (parent) {
+        parent->d->childValueChanged(q, newValue, valueOptions);
+    }
+
+    QVariant prevValue;
+    if (composed && useComposedProperty) {
+        prevValue = value; //???
+        composed->setChildValueChangedEnabled(false);
+        composed->setValue(
+            q, newValue, valueOptions | ValueOption::IgnoreComposedProperty // avoid infinite recursion
+        );
+        composed->setChildValueChangedEnabled(true);
+    }
+    else {
+        prevValue = value;
+    }
+
+    value = newValue;
+
+    if (!parent) { // emit only if parent has not done it
+        emitPropertyChanged(); // called as last step in this method!
+    }
+    return true;
+}
+
+void KProperty::Private::addChild(KProperty *prop)
+{
+    if (!prop) {
+        return;
+    }
+
+    if (!children || qFind(children->begin(), children->end(), prop) == children->end()) { // not in our list
+        if (!children) {
+            children = new QList<KProperty*>();
+        }
+        children->append(prop);
+        prop->d->parent = q;
+    } else {
+        kprWarning() << "property" << name
+                   << ": child property" << prop->name() << "already added";
+        return;
+    }
+}
+
+void KProperty::Private::addSet(KPropertySet *newSet)
+{
+    if (!newSet) {
+        return;
+    }
+
+    if (!set) {//simple case
+        set = newSet;
+        return;
+    }
+    if (set == newSet || (sets && sets->contains(newSet))) {
+        return;
+    }
+    if (!sets) {
+        sets = new QList< QPointer<KPropertySet> >;
+    }
+    sets->append(QPointer<KPropertySet>(newSet));
+}
+
+void KProperty::Private::addRelatedProperty(KProperty *property)
+{
+    if (!relatedProperties)
+        relatedProperties = new QList<KProperty*>();
+
+    if (!relatedProperties->contains(property)) {
+        relatedProperties->append(property);
+    }
+}
+
+void KProperty::Private::emitPropertyChanged()
+{
+    QList< QPointer<KPropertySet> > *realSets = nullptr;
+    if (sets) {
+        realSets = sets;
+    }
+    else if (parent) {
+        realSets = parent->d->sets;
+    }
+    if (realSets) {
+        foreach (QPointer<KPropertySet> s, *realSets) {
+            if (!s.isNull()) { //may be destroyed in the meantime
+                emit s->propertyChangedInternal(*s, *q);
+                emit s->propertyChanged(*s, *q);
             }
         }
-        else {
-            caption = 0;
+    }
+    else {
+        QPointer<KPropertySet> realSet;
+        if (set) {
+            realSet = set;
         }
-        this->captionForDisplaying = captionForDisplaying;
-    }
-
-    ~Private() {
-        delete caption;
-        caption = 0;
-        delete listData;
-        if (children) {
-            qDeleteAll(*children);
-            delete children;
+        else if (parent) {
+            realSet = parent->d->set;
+        }
+        if (!realSet.isNull()) {
+            //if the slot connect with that signal may call set->clear() - that's
+            //the case e.g. at kexi/plugins/{macros|scripting}/* -  this KProperty
+            //may got destroyed ( see KPropertySet::removeProperty(KProperty*) ) while we are
+            //still on it. So, if we try to access ourself/this once the signal
+            //got emitted we may end in a very hard to reproduce crash. So, the
+            //emit should happen as last step in this method!
+            emit realSet->propertyChangedInternal(*realSet, *q);
+            emit realSet->propertyChanged(*realSet, *q);
         }
-        delete relatedProperties;
-        delete composed;
-        delete sets;
-    }
-
-    int type;
-    QByteArray name;
-    QString captionForDisplaying;
-    QString* caption;
-    QString description;
-    QVariant value;
-    QVariant oldValue;
-    /*! The string-to-value correspondence list of the property.*/
-    KPropertyListData* listData;
-    QString icon;
-
-    bool changed;
-    bool storable;
-    bool readOnly;
-    bool visible;
-    int autosync;
-    QMap<QByteArray, QVariant> options;
-
-    KComposedPropertyInterface *composed;
-    //! Flag used to allow composed property to use setValue() without causing recursion
-    bool useComposedProperty;
-
-    //! Used when a single set is assigned for the property
-    QPointer<KPropertySet> set;
-    //! Used when multiple sets are assigned for the property
-    QList< QPointer<KPropertySet> > *sets;
-
-    KProperty  *parent;
-    QList<KProperty*>  *children;
-    //! list of properties with the same name (when intersecting buffers)
-    QList<KProperty*>  *relatedProperties;
-};
-
-/////////////////////////////////////////////////////////////////
-
-KPropertyListData::KPropertyListData(const QStringList& keys_, const QStringList& names_)
-        : names(names_)
-{
-    setKeysAsStringList(keys_);
-}
-
-KPropertyListData::KPropertyListData(const QList<QVariant> keys_, const QStringList& names_)
-        : keys(keys_), names(names_)
-{
-}
-
-KPropertyListData::KPropertyListData()
-{
-}
-
-KPropertyListData::~KPropertyListData()
-{
-}
-
-void KPropertyListData::setKeysAsStringList(const QStringList& list)
-{
-    keys.clear();
-    for (QStringList::ConstIterator it = list.constBegin(); it != list.constEnd(); ++it) {
-        keys.append(*it);
     }
 }
 
-QStringList KPropertyListData::keysAsStringList() const
+void KProperty::Private::childValueChanged(KProperty *child, const QVariant &value,
+                                           KProperty::ValueOptions valueOptions)
 {
-    QStringList result;
-    for (QList<QVariant>::ConstIterator it = keys.constBegin(); it != keys.constEnd(); ++it) {
-        result.append((*it).toString());
+    if (!composed) {
+        return;
     }
-    return result;
+    composed->childValueChangedInternal(
+        child, value,
+        valueOptions | KProperty::ValueOption::IgnoreComposedProperty // avoid infinite recursion
+    );
 }
 
 /////////////////////////////////////////////////////////////////
@@ -144,7 +269,7 @@ QStringList KPropertyListData::keysAsStringList() const
 KProperty::KProperty(const QByteArray &name, const QVariant &value,
                    const QString &caption, const QString &description,
                    int type, KProperty* parent)
-        : d(new KProperty::Private())
+        : d(new KProperty::Private(this))
 {
     d->name = name;
     d->setCaptionForDisplaying(caption);
@@ -156,33 +281,14 @@ KProperty::KProperty(const QByteArray &name, const QVariant &value,
     setType(type);
 
     if (parent)
-        parent->addChild(this);
-    setValue(value, false);
-}
-
-KProperty::KProperty(const QByteArray &name, const QStringList &keys, const QStringList &strings,
-                   const QVariant &value, const QString &caption, const QString &description,
-                   int type, KProperty* parent)
-        : d(new KProperty::Private())
-{
-    d->name = name;
-    d->setCaptionForDisplaying(caption);
-    d->description = description;
-    setListData(keys, strings);
-    if (type == int(Auto)) {
-        type = value.type();
-    }
-    setType(type);
-
-    if (parent)
-        parent->addChild(this);
-    setValue(value, false);
+        parent->d->addChild(this);
+    setValue(value, ValueOption::IgnoreOld);
 }
 
 KProperty::KProperty(const QByteArray &name, KPropertyListData* listData,
                    const QVariant &value, const QString &caption, const QString &description,
                    int type, KProperty* parent)
-        : d(new KProperty::Private())
+        : d(new KProperty::Private(this))
 {
     d->name = name;
     d->setCaptionForDisplaying(caption);
@@ -194,17 +300,17 @@ KProperty::KProperty(const QByteArray &name, KPropertyListData* listData,
     setType(type);
 
     if (parent)
-        parent->addChild(this);
-    setValue(value, false);
+        parent->d->addChild(this);
+    setValue(value, KProperty::ValueOption::IgnoreOld);
 }
 
 KProperty::KProperty()
-        : d(new KProperty::Private())
+        : d(new KProperty::Private(this))
 {
 }
 
 KProperty::KProperty(const KProperty &prop)
-        : d(new KProperty::Private())
+        : d(new KProperty::Private(this))
 {
     *this = prop;
 }
@@ -229,7 +335,7 @@ KProperty::setName(const QByteArray &name)
 QString
 KProperty::caption() const
 {
-    return d->caption ? *d->caption : d->captionForDisplaying;
+    return d->caption.isEmpty() ? d->captionForDisplaying : d->caption;
 }
 
 QString
@@ -273,15 +379,15 @@ KProperty::setType(int type)
 }
 
 QString
-KProperty::icon() const
+KProperty::iconName() const
 {
-    return d->icon;
+    return d->iconName;
 }
 
 void
-KProperty::setIcon(const QString &icon)
+KProperty::setIconName(const QString &name)
 {
-    d->icon = icon;
+    d->iconName = name;
 }
 
 QVariant
@@ -296,112 +402,21 @@ KProperty::oldValue() const
     return d->oldValue;
 }
 
-void
-KProperty::childValueChanged(KProperty *child, const QVariant &value, bool rememberOldValue)
+bool KProperty::setValue(const QVariant &value, ValueOptions options)
 {
-    if (!d->composed)
-        return;
-    d->composed->childValueChangedInternal(child, value, rememberOldValue);
+    return d->setValueInternal(value, options);
 }
 
-//! @return true if @a currentValue and @a value are compatible
-static bool compatibleTypes(const QVariant& currentValue, const QVariant &value)
+void KProperty::setValue(const QVariant &value, bool doNotUseThisOverload, bool doNotUseThisOverload2)
 {
-    if (currentValue.isNull() || value.isNull())
-        return true;
-    const QVariant::Type t = currentValue.type();
-    const QVariant::Type newt = value.type();
-    if (t == newt)
-        return true;
-    if (   (t == QVariant::Int && newt == QVariant::UInt)
-        || (t == QVariant::UInt && newt == QVariant::Int)
-        || (t == QVariant::ByteArray && newt == QVariant::String)
-        || (t == QVariant::String && newt == QVariant::ByteArray)
-        || (t == QVariant::ULongLong && newt == QVariant::LongLong)
-        || (t == QVariant::LongLong && newt == QVariant::ULongLong))
-    {
-        return true;
-    }
-    return false;
+    Q_UNUSED(value);
+    Q_UNUSED(doNotUseThisOverload);
+    Q_UNUSED(doNotUseThisOverload2);
 }
 
-void KProperty::setValue(const QVariant &value, bool rememberOldValue, bool useComposedProperty)
+bool KProperty::valueEqualsTo(const QVariant &value, ValueOptions valueOptions) const
 {
-    if (d->name.isEmpty()) {
-        kprWarning() << "COULD NOT SET value to a null property";
-        return;
-    }
-    const QVariant currentValue = this->value();
-    if (!compatibleTypes(currentValue, value)) {
-        kprWarning() << "INCOMPATIBLE TYPES! old=" << currentValue << "new=" << value;
-    }
-
-    //1. Check if the value should be changed
-    bool ch;
-    const QVariant::Type t = currentValue.type();
-    const QVariant::Type newt = value.type();
-    if (   t == QVariant::DateTime
-        || t == QVariant::Time)
-    {
-        //for date and datetime types: compare with strings, because there
-        //can be miliseconds difference
-        ch = (currentValue.toString() != value.toString());
-    }
-    else if (t == QVariant::String || t == QVariant::ByteArray) {
-        //property is changed for string type,
-        //if one of value is empty and other isn't..
-        ch = ((currentValue.toString().isEmpty() != value.toString().isEmpty())
-              //..or both are not empty and values differ
-              || (!currentValue.toString().isEmpty() && !value.toString().isEmpty() && currentValue != value));
-    }
-    else if (t == QVariant::Double) {
-        const double factor = 1.0 / option("step", KPROPERTY_DEFAULT_DOUBLE_VALUE_STEP).toDouble();
-        //kprDebug()
-        //    << "double compared:" << currentValue.toDouble() << value.toDouble()
-        //    << ":" << static_cast<qlonglong>(currentValue.toDouble() * factor) << static_cast<qlonglong>(value.toDouble() * factor);
-        ch = static_cast<qlonglong>(currentValue.toDouble() * factor) != static_cast<qlonglong>(value.toDouble() * factor);
-    } else if (t == QVariant::Invalid && newt == QVariant::Invalid) {
-        ch = false;
-    } else if (d->composed && useComposedProperty) {
-        ch = !d->composed->valuesEqual(currentValue, value);
-    }
-    else {
-        ch = (currentValue != value);
-    }
-
-    if (!ch)
-        return;
-
-    //2. Then change it, and store old value if necessary
-    if (rememberOldValue) {
-        if (!d->changed)
-            d->oldValue = currentValue;
-        d->changed = true;
-    }
-    else {
-        d->oldValue = QVariant(); // clear old value
-        d->changed = false;
-    }
-    if (d->parent) {
-        d->parent->childValueChanged(this, value, rememberOldValue);
-    }
-
-    QVariant prevValue;
-    if (d->composed && useComposedProperty) {
-        prevValue = currentValue; //???
-        d->composed->setChildValueChangedEnabled(false);
-        d->composed->setValue(this, value, rememberOldValue);
-        d->composed->setChildValueChangedEnabled(true);
-    }
-    else {
-        prevValue = currentValue;
-    }
-
-    d->value = value;
-
-    if (!d->parent) { // emit only if parent has not done it
-        emitPropertyChanged(); // called as last step in this method!
-    }
+    return !d->valueDiffersInternal(value, valueOptions);
 }
 
 void
@@ -409,9 +424,10 @@ KProperty::resetValue()
 {
     d->changed = false;
     bool cleared = false;
-    if (d->set)
-        d->set->informAboutClearing(cleared); //inform me about possibly clearing the property sets
-    setValue(oldValue(), false);
+    if (d->set) {
+        KPropertySetPrivate::d(d->set)->informAboutClearing(&cleared); //inform me about possibly clearing the property sets
+    }
+    setValue(oldValue(), ValueOption::IgnoreOld);
     if (cleared)
         return; //property set has been cleared: no further actions make sense as 'this' is dead
 
@@ -461,13 +477,28 @@ KProperty::isNull() const
 bool
 KProperty::isModified() const
 {
-    return d->changed;
+    if (d->changed) {
+        return true;
+    }
+    if (d->children) {
+        for (const KProperty* p : *d->children) {
+            if (p->isModified()) {
+                return true;
+            }
+        }
+    }
+    return false;
 }
 
 void
 KProperty::clearModifiedFlag()
 {
     d->changed = false;
+    if (d->children) {
+        for (KProperty* p : *d->children) {
+            p->clearModifiedFlag();
+        }
+    }
 }
 
 bool
@@ -494,16 +525,15 @@ KProperty::setVisible(bool visible)
     d->visible = visible;
 }
 
-int
-KProperty::autoSync() const
+KProperty::ValueSyncPolicy KProperty::valueSyncPolicy() const
 {
-    return d->autosync;
+    return d->valueSyncPolicy;
 }
 
 void
-KProperty::setAutoSync(int sync)
+KProperty::setValueSyncPolicy(KProperty::ValueSyncPolicy policy)
 {
-    d->autosync = sync;
+    d->valueSyncPolicy = policy;
 }
 
 bool
@@ -521,21 +551,22 @@ KProperty::setStorable(bool storable)
 void
 KProperty::setOption(const char* name, const QVariant& val)
 {
-    d->options[name] = val;
+    if (val.isNull()) {
+        d->options.remove(name);
+    } else {
+        d->options[name] = val;
+    }
 }
 
-QVariant
-KProperty::option(const char* name, const QVariant& defaultValue) const
+QVariant KProperty::option(const char* name, const QVariant& defaultValue) const
 {
-    if (d->options.contains(name))
-        return d->options[name];
-    return defaultValue;
+    return d->option(name, defaultValue);
 }
 
 bool
 KProperty::hasOptions() const
 {
-    return !d->options.isEmpty();
+    return !d->options.isEmpty() || (d->parent && d->parent->hasOptions());
 }
 
 /////////////////////////////////////////////////////////////////
@@ -554,21 +585,21 @@ KProperty::operator= (const KProperty & property)
         return *this;
 
     delete d->listData;
-    d->listData = 0;
+    d->listData = nullptr;
     delete d->children;
-    d->children = 0;
+    d->children = nullptr;
     delete d->relatedProperties;
-    d->relatedProperties = 0;
+    d->relatedProperties = nullptr;
     delete d->composed;
-    d->composed = 0;
+    d->composed = nullptr;
 
     d->name = property.d->name;
     d->setCaptionForDisplaying(property.captionForDisplaying());
     d->description = property.d->description;
     d->type = property.d->type;
 
-    d->icon = property.d->icon;
-    d->autosync = property.d->autosync;
+    d->iconName = property.d->iconName;
+    d->valueSyncPolicy = property.d->valueSyncPolicy;
     d->visible = property.d->visible;
     d->storable = property.d->storable;
     d->readOnly = property.d->readOnly;
@@ -590,7 +621,7 @@ KProperty::operator= (const KProperty & property)
             QList<KProperty*>::ConstIterator endIt = property.d->children->constEnd();
             for (QList<KProperty*>::ConstIterator it = property.d->children->constBegin(); it != endIt; ++it) {
                 KProperty *child = new KProperty(*(*it));
-                addChild(child);
+                d->addChild(child);
             }
         }
     }
@@ -611,6 +642,11 @@ KProperty::operator ==(const KProperty &prop) const
     return ((d->name == prop.d->name) && (value() == prop.value()));
 }
 
+bool KProperty::operator!=(const KProperty &prop) const
+{
+    return !operator==(prop);
+}
+
 /////////////////////////////////////////////////////////////////
 
 const QList<KProperty*>*
@@ -627,7 +663,7 @@ KProperty::child(const QByteArray &name)
         if ((*it)->name() == name)
             return *it;
     }
-    return 0;
+    return nullptr;
 }
 
 KProperty*
@@ -636,60 +672,6 @@ KProperty::parent() const
     return d->parent;
 }
 
-void
-KProperty::addChild(KProperty *prop)
-{
-    if (!prop)
-        return;
-
-    if (!d->children || qFind(d->children->begin(), d->children->end(), prop) == d->children->end()) { // not in our list
-        if (!d->children)
-            d->children = new QList<KProperty*>();
-        d->children->append(prop);
-        prop->d->parent = this;
-    } else {
-        kprWarning() << "property" << name()
-                   << ": child property" << prop->name() << "already added";
-        return;
-    }
-}
-
-void
-KProperty::addSet(KPropertySet *set)
-{
-    if (!set)
-        return;
-
-    if (!d->set) {//simple case
-        d->set = set;
-        return;
-    }
-    if (d->set == set || d->sets->contains(set))
-        return;
-    if (!d->sets) {
-        d->sets = new QList< QPointer<KPropertySet> >;
-    }
-
-    d->sets->append(QPointer<KPropertySet>(set));
-}
-
-const QList<KProperty*>*
-KProperty::related() const
-{
-    return d->relatedProperties;
-}
-
-void
-KProperty::addRelatedProperty(KProperty *property)
-{
-    if (!d->relatedProperties)
-        d->relatedProperties = new QList<KProperty*>();
-
-    QList<KProperty*>::iterator it = qFind(d->relatedProperties->begin(), d->relatedProperties->end(), property);
-    if (it == d->relatedProperties->end()) // not in our list
-        d->relatedProperties->append(property);
-}
-
 KComposedPropertyInterface* KProperty::composedProperty() const
 {
     return d->composed;
@@ -716,63 +698,17 @@ void Property::setSortingKey(int key)
 }
 #endif
 
-void KProperty::emitPropertyChanged()
-{
-    QList< QPointer<KPropertySet> > *sets = 0;
-    if (d->sets) {
-        sets = d->sets;
-    }
-    else if (d->parent) {
-        sets = d->parent->d->sets;
-    }
-    if (sets) {
-        foreach (QPointer<KPropertySet> s, *sets) {
-            if (!s.isNull()) { //may be destroyed in the meantime
-                emit s->propertyChangedInternal(*s, *this);
-                emit s->propertyChanged(*s, *this);
-            }
-        }
-    }
-    else {
-        QPointer<KPropertySet> set;
-        set = d->set;
-        if (d->set) {
-            set = d->set;
-        }
-        else if (d->parent) {
-            set = d->parent->d->set;
-        }
-        if (!set.isNull()) {
-            //if the slot connect with that signal may call set->clear() - that's
-            //the case e.g. at kexi/plugins/{macros|scripting}/* -  this Property
-            //may got destroyed ( see Set::removeProperty(Property*) ) while we are
-            //still on it. So, if we try to access ourself/this once the signal
-            //got emitted we may end in a very hard to reproduce crash. So, the
-            //emit should happen as last step in this method!
-            emit set->propertyChangedInternal(*set, *this);
-            emit set->propertyChanged(*set, *this);
-        }
-    }
-}
-
-QMap<QByteArray, QVariant> KProperty::options() const
-{
-    return d->options;
-}
-
 /////////////////////////////////////////////////////////////////
 
-void KProperty::debug() const
-{
-    kprDebug() << *this;
-}
-
 KPROPERTYCORE_EXPORT QDebug operator<<(QDebug dbg, const KProperty &p)
 {
     dbg.nospace() << "KProperty("
         << "NAME=" << p.name();
-    if (!p.caption().isEmpty()) {
-        dbg.nospace() << " CAPTION=" << p.caption();
+    if (!p.captionForDisplaying().isEmpty()) {
+        dbg.nospace() << " CAPTION_FOR_DISPLAYING=" << p.captionForDisplaying();
+        if (p.captionForDisplaying() != p.caption()) {
+            dbg.nospace() << " CAPTION=" << p.caption();
+        }
     }
     if (!p.description().isEmpty()) {
         dbg.nospace() << " DESC=" << p.description();
@@ -797,8 +733,8 @@ KPROPERTYCORE_EXPORT QDebug operator<<(QDebug dbg, const KProperty &p)
 //! @todo children...
 
     if (p.hasOptions()) {
-        dbg.nospace() << " OPTIONS(" << p.options().count() << "): [";
-        QList<QByteArray> optionKeys( p.options().keys() );
+        dbg.nospace() << " OPTIONS(" << p.d->options.count() << "): [";
+        QList<QByteArray> optionKeys( p.d->options.keys() );
         qSort(optionKeys);
         bool first = true;
         foreach (const QByteArray& key, optionKeys) {
