@@ -1,7 +1,7 @@
 /* This file is part of the KDE project
    Copyright (C) 2004 Cedric Pasteur <cedric.pasteur@free.fr>
    Copyright (C) 2004 Alexander Dymo <cloudtemple@mskat.net>
-   Copyright (C) 2004-2016 Jarosław Staniek <staniek@kde.org>
+   Copyright (C) 2004-2017 Jarosław Staniek <staniek@kde.org>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -27,36 +27,33 @@
 #include <QByteArray>
 #include <QDebug>
 
+#include <cmath>
+#include <limits>
+
 #include "kpropertycore_export.h"
 
 class KComposedPropertyInterface;
+class KPropertyListData;
 class KPropertySet;
+class KPropertySetPrivate;
 
-/*! Data container for properties of list type. */
-class KPROPERTYCORE_EXPORT KPropertyListData
-{
-public:
-    /*! Data container for list-value property.
-     We will be able to choose an item from this list. */
-    KPropertyListData(const QStringList& keys_, const QStringList& names_);
-    KPropertyListData(const QList<QVariant> keys_, const QStringList& names_);
-    KPropertyListData();
-    ~KPropertyListData();
+/**
+ * @brief Minimum double value working precisely.
+ *
+ * @since 3.1
+ */
+#define KPROPERTY_MIN_PRECISE_DOUBLE (-pow(2, std::numeric_limits<double>::digits))
 
-    void setKeysAsStringList(const QStringList& list);
-    QStringList keysAsStringList() const;
-
-    /*! The string list containing all possible keys for this property
-     or NULL if this is not a property of type 'list'. The values in this list are ordered,
-     so the first key element is associated with first element from
-     the 'names' list, and so on. */
-    QList<QVariant> keys;
-
-//! @todo what about using QValueList<QVariant> here too?
-    /*! The list of translated names that will be visible on the screen.
-     First value is referenced by first key, and so on. */
-    QStringList names;
-};
+/**
+ * @brief Minimum double value working precisely
+ *
+ * Editor for double values (spin box) has localized contents and its code supports just this maximum.
+ * For a 64-bit machine it's 2**53.
+ * See also https://phabricator.kde.org/D5419#inline-22329
+ *
+ * @since 3.1
+ */
+#define KPROPERTY_MAX_PRECISE_DOUBLE (pow(2, std::numeric_limits<double>::digits))
 
 /*! \brief The base class representing a single property
 
@@ -77,17 +74,14 @@ public:
   KProperty *property = new KProperty(name, value, caption, description);
   // name is a QByteArray, value is whatever type QVariant supports
 
-  // Creating a valueFromList property matching keys with names:
-  QStringList keys, names;
-  keys << "one" << "two" << "three"; // possible values of the property
-  // Names (possibly translated) shown in the editor instead of the keys
-  names << tr("One") << tr("Two") << tr("Three");
-  property = new KProperty(name, keys, names, "two", caption);
+  // Creating a property of type ValueFromList matching keys with names:
+  QStringList keys({"one", "two", "three"}); // possible values of the property
+  QStringList names({tr("One"), tr("Two"), tr("Three")}); // Names (possibly translated) shown in
+                                                          // the editor instead of the keys
+  property = new KProperty(name, new KPropertyListData(keys, names), "two", caption);
 
-  // Creating a valueFromList property matching QVariant keys with names:
-  QValueList<QVariant> variantKeys;
-  variantKeys << 1 << 2 << 3;
-  KPropertyListData *listData = new KPropertyListData(variantKeys, names);
+  // Creating a property of type ValueFromList matching variant keys with names:
+  KPropertyListData *listData = new KPropertyListData({1.1, tr("One")}, {2.5, tr("Two")}, {3., tr("Three")}});
   propertySet->addProperty(new KProperty("List", listData, "otheritem", "List"));
   @endcode
 
@@ -165,26 +159,23 @@ public:
         UserDefined = 4000            /**<plugin defined properties should start here*/
     };
 
-    /*! Constructs a null property. */
+    /**
+     * Constructs a null property.
+     * Null properties have empty names and captions and Invalid types.
+     */
     KProperty();
 
-    /*! Constructs property of simple type.
-     If \a caption contains newline characters, caption() will return \a caption with substituted
-     these with spaces. captionForDisplaying() is used to get original caption text usable
-     (with newline, if any) for displaying within a property editor. */
+    /*! Constructs property of a simple type. */
     explicit KProperty(const QByteArray &name, const QVariant &value = QVariant(),
              const QString &caption = QString(), const QString &description = QString(),
              int type = Auto, KProperty* parent = nullptr);
 
-    /*! Constructs property of \ref ValueFromList type. */
-    KProperty(const QByteArray &name, const QStringList &keys, const QStringList &strings,
-             const QVariant &value = QVariant(),
-             const QString &caption = QString(), const QString &description = QString(),
-             int type = ValueFromList, KProperty* parent = nullptr);
-
-    /*! Constructs property of \ref ValueFromList type.
-     This is overload of the above ctor added for convenience. */
-    KProperty(const QByteArray &name, KPropertyListData* listData,
+    /**
+     * @brief Constructs property of ValueFromList type
+     *
+     * Ownership of @a listData is passed to the property object.
+     */
+    KProperty(const QByteArray &name, KPropertyListData *listData,
              const QVariant &value = QVariant(),
              const QString &caption = QString(), const QString &description = QString(),
              int type = ValueFromList, KProperty* parent = nullptr);
@@ -194,23 +185,36 @@ public:
 
     ~KProperty();
 
-    /*! \return the internal name of the property (that's used in List).*/
+    /**
+     * @return name of the property
+     * @note empty name means a null property
+     */
     QByteArray name() const;
 
-    /*! Sets the internal name of the property.*/
+    /**
+     * Sets name of the property
+     * @note empty name means a null property
+     */
     void setName(const QByteArray &name);
 
-    /*! \return the caption of the property.*/
+    /*! \return the caption of the property. Does not contain newline characters. Can be empty. */
     QString caption() const;
 
+    /*! \return the caption of the property or name() if the caption is empty. */
+    inline QString captionOrName() const
+    {
+        return caption().isEmpty() ? QString::fromLatin1(name()) : caption();
+    }
+
     /*! \return the caption text of the property for displaying.
      It is similar to caption() but if the property caption contains newline characters,
      these are not substituted with spaces. */
     QString captionForDisplaying() const;
 
     /*! Sets the name of the property. If the caption contains newline characters,
-     these are replaced by spaces. captionForDisplaying() can be used to access the original caption
-     text passed here.*/
+     these are substituted with spaces.
+     @see captionForDisplaying
+    */
     void setCaption(const QString &caption);
 
     /*! \return the description of the property.*/
@@ -228,14 +232,38 @@ public:
     /*! \return the value of the property.*/
     QVariant value() const;
 
-    /*! Gets the previous property value.*/
+    /*! Returns the previous property value if it was set in setValue(). */
     QVariant oldValue() const;
 
-    void childValueChanged(KProperty *child, const QVariant &value, bool rememberOldValue);
+    //! Options that influence how values are handled in setValue() and valueEqualsTo()
+    //! @since 3.1
+    enum class ValueOption {
+        None = 0,        //!< No options, that is 1. old value is remembered before setting a new one;
+                         //!< 2. composed properties are considered while comparing old and new value
+        IgnoreOld = 1,   //!< Do not remember the old value before setting a new one
+        IgnoreComposedProperty = 2 //!< Do not use composed property when comparing values
+    };
+    Q_DECLARE_FLAGS(ValueOptions, ValueOption)
 
-    //! Sets value of the property to @a value
-    //! @todo 4.0 BCI: replace with bool setValue(QVariant, ValueOptions)
-    void setValue(const QVariant &value, bool rememberOldValue = true, bool useComposedProperty = true);
+    /**
+     * @brief Sets value of the property.
+     * @param value New value
+     * @param options Options for the value setting.
+     * @return @c true if the value has been changed and @c false if the @a value was the same
+     *         as previous one so it was not changed of if this property is null.
+     */
+    bool setValue(const QVariant &value, ValueOptions options = ValueOptions());
+
+    /**
+     * @return true if value of this property is equal to specified value
+     *
+     * Takes type into account.
+     * @param value Value to compare.
+     * @param valueOptions Options to use when comparing.
+     *                     Only the @c None and IgnoreComposedProperties are supported.
+     * @since 3.1
+     */
+    bool valueEqualsTo(const QVariant &value, ValueOptions valueOptions = ValueOptions()) const;
 
     /*! Resets the value of the property to the old value.
      @see oldValue() */
@@ -254,12 +282,12 @@ public:
      This is overload of the above ctor added for convenience. */
     void setListData(const QStringList &keys, const QStringList &names);
 
-    /*! Sets icon by \a name for this property. Icons are optional and are used e.g.
+    /*! Sets icon name to \a name for this property. Icons are optional and are used e.g.
      in property editor - displayed at the left hand. */
-    void setIcon(const QString &icon);
+    void setIconName(const QString &name);
 
-    /*! \return property icon's string. Can be empty. */
-    QString icon() const;
+    /*! \return property icon's name. Can be empty. */
+    QString iconName() const;
 
     /*! \return a list of all children for this property, or NULL of there
      is no children for this property */
@@ -278,13 +306,24 @@ public:
     /*! Sets composed property \a prop for this property. */
     void setComposedProperty(KComposedPropertyInterface *prop);
 
-    /*! \return true if this property is null. Null properties have empty names. */
+    /*! \return true if this property is null. Property is null if it has empty name. */
     bool isNull() const;
 
-    //! \return true if this property value is changed.
+    /**
+     * @brief Return @c true if value of this property or value of any child property is modified.
+     *
+     * @see clearModifiedFlag()
+     */
     bool isModified() const;
 
-    //! Clears "modified" flag, so isModified() will return false.
+    /**
+     * @brief Clears the "modified" flag for this property and all its child properties.
+     *
+     * After calling this method isModified() returs false for the property and all child
+     * properties.
+     *
+     * @see isModified()
+     */
     void clearModifiedFlag();
 
     /*! \return true if the property is read-only when used in a property editor.
@@ -313,33 +352,62 @@ public:
     /*! Sets "storable" flag for this property. @see isStorable() */
     void setStorable(bool storable);
 
-    /*! \return 1 if the property should be synced automatically in the property editor
-    as soon as editor contents change (e.g. when the user types text).
-    If autoSync() == 0, property value will be updated when the user presses Enter
-    or when another editor gets the focus.
-    Property follows property editor's global rule if autoSync() !=0 and !=1 (the default).
-    */
-    int autoSync() const;
+    //! Synchronization policy for property values
+    //! @since 3.1
+    enum class ValueSyncPolicy {
+        Editor,   //!< Allow to synchronize by the property editor using its valueSync setting (default)
+        FocusOut, //!< Synchronize the value when focus is out of the editor widget for this property
+                  //!< or when the user presses the Enter key
+        Auto      //!< Synchronize automatically as soon as the editor widget for this property signals
+                  //! (using commitData) that the value has been changed, e.g. when the user types
+                  //! another letter in a text box
+    };
 
-    /*! If \a sync is 1, the property will be synced automatically in the property editor
-    as soon as editor's contents change (e.g. when the user types text).
-    If \a sync is 0, property value will be updated when the user presses
-    Enter or when another editor gets the focus.
-    Property follows property editor's global rule if sync !=0 and !=1 (the default).
-    */
-    void setAutoSync(int sync);
+    //! @return synchronization policy for property values of this property
+    //! @since 3.1
+    ValueSyncPolicy valueSyncPolicy() const;
+
+    //! Sets synchronization policy for property values of this property
+    //! See ValueSyncPolicy for details.
+    //! @since 3.1
+    void setValueSyncPolicy(ValueSyncPolicy policy);
 
     /*! Sets value \a val for option \a name.
      Options are used to override default settings of individual properties.
-     This is most visible in property editor widget.
+     They are most visible in property editor widgets. Option is set if it is not null.
+     This means that empty string can be still a valid value.
+     To unset given option, call setOption() with a null QVariant value.
 
     Currently supported options are:
-    <ul><li> min: integer value describing minimum value for properties of integer and
-             double types. The default is 0.</li>
-    <li> minValueText: user-visible translated string to be displayed in editor for integer
-         type when minimum is set for the property.
+    <ul>
+    <li> min: value describing minimum value for properties of integer, double,
+         date, date/time and time types. Default is 0 for double and unsigned integer types,
+         -INT_MAX for signed integer type.
+         Defaults for date, date/time and time types are specified in documentation
+         of QDateEdit::minimumDate, QDateTimeEdit::minimumDateTime and QTime::minimumTime, respectively.
+         The value specified for this option is accepted if:
+         - it is not larger than the value of the "max" option
+         - it is not smaller than KPROPERTY_MIN_PRECISE_DOUBLE (for double type)
+         - it is not smaller than -INT_MAX (for integer type)
+         - it is not smaller than 0 (for unsigned integer type).
+         </li>
+    <li> minValueText: user-visible translated string to be displayed in editor for integer,
+         double, date, date/time and time types when the value is equal to the value of
+         "min" option.
+         The value specified for this option is accepted if min option is supported for given type
+         and is specified.
          @see QAbstractSpinBox::specialValueText</li>
-    <li> max: integer describing minimum value for properties of integer type. Default is 0xffff.</li>
+    <li> max: value describing minimum value for properties of integer type.
+         Default is KPROPERTY_MAX_PRECISE_DOUBLE for double type (maximum precise value)
+         and INT_MAX for integer type.
+         Defaults for date, date/time and time types are specified in documentation
+         of QDateEdit::maximumDate, QDateTimeEdit::maximumDateTime and QTime::maximumTime, respectively.
+         The value is ignored if it is smaller than the value of "min" option. </li>
+         The value specified for this option is accepted if:
+         - it is not smaller than the value of the "min" option
+         - it is not larger than KPROPERTY_MAX_PRECISE_DOUBLE (for double type)
+         - it is not larger than INT_MAX (for integer and unsigned integer type).
+         </li>
     <li> precision: integer value >= 0 describing the number of decimals after the decimal
          point for double type. Default value is 2.
          @see QDoubleSpinBox::decimals</li>
@@ -360,28 +428,57 @@ public:
          @c false.</li>
     <li> extraValueAllowed: boolean value, if @c true the user is able to manually add extra
          values to a combobox.</li>
-    <li> fileMode: string value that describes what objects may select in the file dialog
-         of the url editor:
+    <li> fileMode: string value that describes types of objects that can be selected by the url
+         editor:
          <ul>
-         <li>"dirsOnly": only support and display directories;
-             @see QFileDialog::ShowDirsOnly QFileDialog::Directory</li>
-         <li>"existingFile": only allow to select one existing file;
-             @see QFileDialog::ExistingFile</li>
-         <li>Any other value: any file is supported, whether it exists or not
-             @see QFileDialog::AnyFile</li>
+         <li>"dirsOnly": only display and allow to select existing directories;
+             @see QFileDialog::getExistingDirectoryUrl()</li>
+         <li>"existingFile": only allow to select one existing file for opening, i.e. confirmation
+             of overwriting is not performed;
+             @see QFileDialog::getOpenFileUrl()</li>
+         <li>Any other value: any file is supported, whether it exists or not; if the file exists,
+             "confirmOverwrites" option is honored; this mode is the only one supporting non-file
+             protocols such as ftp or http; to use them user has to enter them explicitly,
+             file protocol is still the default
+             @see QFileDialog::getSaveFileUrl()</li>
          </ul>
+         @note Empty URLs are always allowed.
          </li>
-    <li> confirmOverwrites: boolean value, if @c true, user will be asked for confirmation
-         of file overwriting in the url editor. @c false by default.
+    <li> confirmOverwrites: boolean value supported by the url editor; if @c true and the "fileMode"
+         option is not equal to "existingFile" nor "dirsOnly" user will be asked for confirmation
+         of file overwriting if selected file exists. @c false by default.
          @note The line edit does not validate the content.</li>
+    <li> multiLine: boolean value used for string type. If @c true, a multi-line
+         QPlainTextEdit-based widget is used for editor; otherwise a single-line QLineEdit
+         widget is used. @c false by default. Added in version 3.1.</li>
+    <li>prefix: string to display before the value, e.g. '$'. Supported for double and integer
+        types and composed types based on double and integer types (Point*, Size*, Rect*).
+        @see QDoubleSpinBox::prefix QSpinBox::prefix</li>
+    <li>suffix: string to display after the value, e.g. unit such as 'mm'.
+        Supported for double and integer types and composed types based on double and
+        integer types (Point*, Size*, Rect*). Note that only display is affected, value
+        is not converted to any unit.
+        @see QDoubleSpinBox::suffix QSpinBox::suffix</li>
     </ul>*/
     void setOption(const char* name, const QVariant& val);
 
-    /*! \return a value for option \a name or @a defaultValue value if there is no such option set.
-     @see setOption */
+    /*! @brief Returns value of given option
+     * Option is set if returned value is not null.
+     * If there is no option for @a name in given property and parent property is present (see parent()),
+     * parent property is checked. If the parent property offers the option, the value
+     * is returned. If it is not present there, @a defaultValue value is returned.
+     * Looking at parent property is available since 3.1.
+     * @note The lookup is performed recursively, first in parent, then grand parent, etc.
+     * @see setOption
+     */
     QVariant option(const char* name, const QVariant& defaultValue = QVariant()) const;
 
-    /*! \return true if at least one option is defined for this property. */
+    /*! @brief Returns @c true if at least one option is specified for this property
+     * If there are no options defined @c true can be still returned if parent property
+     * is present and it has at least one option specified.
+     * Looking at parent property is available since 3.1.
+     * @note The lookup is performed recursively, first in parent, then grand parent, etc.
+     */
     bool hasOptions() const;
 
     /*! Equivalent to setValue(const QVariant &) */
@@ -390,8 +487,20 @@ public:
     /*! Assigns a deep copy of all attributes of \a property to this property. */
     KProperty& operator= (const KProperty &property);
 
-    /*! Compares two properties.*/
-    bool operator ==(const KProperty &prop) const;
+    /**
+     * @return @c true if the property is equal to @a prop; otherwise returns @c false.
+     * Two properties are equal if they have the same name and type.
+     * @note All null properties are equal
+     * @todo Compare properties deeper?
+     */
+    bool operator==(const KProperty &prop) const;
+
+    /**
+     * @return @c true if the property is different from @a prop; otherwise returns @c false.
+     * Two properties are different if they have different names or types.
+     * @since 3.1
+     */
+    bool operator!=(const KProperty &prop) const;
 
 #if 0
     /*! \return a key used for sorting.
@@ -401,47 +510,22 @@ public:
     int sortingKey() const;
 #endif
 
-protected:
-    /*! Adds \a prop as a child of this property.
-     The children will be owned by this property. */
-    void addChild(KProperty *prop);
+private:
+    //! Added only to help porting old code. Use public setValue() methods.
+    void setValue(const QVariant &value, bool a1, bool a2 = true);
 
-    /*! Adds \a set to this property. */
-    void addSet(KPropertySet *set);
-
-#if 0
-    /*! Sets a key used for sorting. */
-    void setSortingKey(int key);
-#endif
-
-    /*! \return a list of related properties for this property. */
-    const QList<KProperty*>* related() const;
-
-    /*! Adds related property for this property. */
-    void addRelatedProperty(KProperty *property);
-
-    /*! This method emits the \a KPropertySet::propertyChanged() signal.
-    KProperty::setValue() calls this method if the value has been changed. */
-    void emitPropertyChanged();
-
-    /*! Outputs debug string for this property to the standard debug output. */
-    void debug() const;
-
-    /*! Used by debug operator <<. */
-    QMap<QByteArray, QVariant> options() const;
-
-    //! @internal
     class Private;
-    friend class Private;
     Private * const d;
 
     friend class KPropertySet;
-    friend class KPropertyBuffer;
-    friend class KComposedPropertyInterface;
+    friend class KPropertySetPrivate;
+    friend class KPropertySetBuffer;
     friend KPROPERTYCORE_EXPORT QDebug operator<<(QDebug dbg, const KProperty &p);
 };
 
 //! qDebug() stream operator. Writes property @a p to the debug output in a nicely formatted way.
 KPROPERTYCORE_EXPORT QDebug operator<<(QDebug dbg, const KProperty &p);
 
+Q_DECLARE_OPERATORS_FOR_FLAGS(KProperty::ValueOptions)
+
 #endif
