@@ -1,5 +1,5 @@
 /* This file is part of the KDE project
-   Copyright (C) 2008-2016 Jarosław Staniek <staniek@kde.org>
+   Copyright (C) 2008-2018 Jarosław Staniek <staniek@kde.org>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -18,7 +18,7 @@
 */
 
 #include "KPropertyEditorView.h"
-#include "KPropertyEditorDataModel.h"
+#include "KPropertyEditorDataModel_p.h"
 #include "KProperty.h"
 #include "KPropertySet.h"
 #include "KPropertyWidgetsFactory.h"
@@ -67,10 +67,12 @@ public:
 };
 #endif
 
-static bool computeAutoSync(const KProperty *property, bool defaultAutoSync)
+static bool effectiveValueSyncPolicy(const KProperty *property, bool defaultValue)
 {
-    return (property->autoSync() != 0 && property->autoSync() != 1) ?
-                defaultAutoSync : (property->autoSync() != 0);
+    if (property->valueSyncPolicy() == KProperty::ValueSyncPolicy::Editor) {
+        return defaultValue;
+    }
+    return property->valueSyncPolicy() == KProperty::ValueSyncPolicy::Auto;
 }
 
 //----------
@@ -79,13 +81,13 @@ class ItemDelegate : public QItemDelegate
 {
 public:
     explicit ItemDelegate(KPropertyEditorView *parent);
-    virtual ~ItemDelegate();
-    virtual void paint(QPainter *painter,
-        const QStyleOptionViewItem &option, const QModelIndex &index) const;
-    virtual QSize sizeHint(const QStyleOptionViewItem &option,
-        const QModelIndex &index) const;
-    virtual QWidget * createEditor(QWidget *parent,
-        const QStyleOptionViewItem & option, const QModelIndex & index ) const;
+    ~ItemDelegate() override;
+    void paint(QPainter *painter,
+        const QStyleOptionViewItem &option, const QModelIndex &index) const override;
+    QSize sizeHint(const QStyleOptionViewItem &option,
+        const QModelIndex &index) const override;
+    QWidget * createEditor(QWidget *parent,
+        const QStyleOptionViewItem &option, const QModelIndex &index) const override;
     mutable QPointer<QWidget> m_currentEditor;
 };
 
@@ -116,20 +118,21 @@ void ItemDelegate::paint(QPainter *painter,
                          const QModelIndex &index) const
 {
     QStyleOptionViewItem alteredOption(option);
-    const QColor gridLineColor(qobject_cast<KPropertyEditorView*>(parent())->gridLineColor());
-    if (gridLineColor.isValid()) {
-        alteredOption.rect.setTop(alteredOption.rect.top() + 1);
-    }
-    const KPropertyUtils::PainterSaver saver(painter);
-    QRect r(option.rect);
+    const KPropertyUtilsPrivate::PainterSaver saver(painter);
     const KPropertyEditorDataModel *editorModel = qobject_cast<const KPropertyEditorDataModel*>(index.model());
     if (!editorModel) {
         return;
     }
+
+    QRect r(option.rect);
     bool modified = false;
+    const QColor gridLineColor(qobject_cast<KPropertyEditorView*>(parent())->gridLineColor());
+    if (gridLineColor.isValid()) {
+        alteredOption.rect.setTop(alteredOption.rect.top() + 1);
+    }
     if (index.column()==0) {
         r.setWidth(r.width() - 1);
-        r.setLeft(0);
+        r.setLeft(-1); // to avoid displaying double left border
 
         QVariant modifiedVariant( editorModel->data(index, KPropertyEditorDataModel::PropertyModifiedRole) );
         if (modifiedVariant.isValid() && modifiedVariant.toBool()) {
@@ -149,44 +152,51 @@ void ItemDelegate::paint(QPainter *painter,
         alteredOption.rect.setRight( alteredOption.rect.right() - iconSize * 1 );
     }
 
-    KProperty *property = editorModel->propertyForIndex(index);
-    const int t = typeForProperty( property );
-    bool useQItemDelegatePaint = true; // ValueDisplayInterface is used by default
-    if (index.column() == 1 && KPropertyWidgetsPluginManager::self()->paint(t, painter, alteredOption, index)) {
-        useQItemDelegatePaint = false;
-    }
-    if (useQItemDelegatePaint) {
-        QItemDelegate::paint(painter, alteredOption, index);
-    }
+    const bool isGroupHeader(editorModel->data(index, KPropertyEditorDataModel::PropertyGroupRole).toBool());
+    if (!isGroupHeader) {
+        KProperty *property = editorModel->propertyForIndex(index);
+        const int t = typeForProperty( property );
+        bool useQItemDelegatePaint = true; // ValueDisplayInterface is used by default
+        if (index.column() == 1 && KPropertyWidgetsPluginManager::self()->paint(t, painter, alteredOption, index)) {
+            useQItemDelegatePaint = false;
+        }
+        if (useQItemDelegatePaint) {
+            QItemDelegate::paint(painter, alteredOption, index);
+        }
 
-    if (modified) {
-        alteredOption.rect.setRight( alteredOption.rect.right() - iconSize * 3 / 2 );
-        int y1 = alteredOption.rect.top();
-        QLinearGradient grad(x2 - iconSize * 2, y1, x2 - iconSize / 2, y1);
-        QColor color(
-            alteredOption.palette.color(
-                (alteredOption.state & QStyle::State_Selected) ? QPalette::Highlight : QPalette::Base ));
-        color.setAlpha(0);
-        grad.setColorAt(0.0, color);
-        color.setAlpha(255);
-        grad.setColorAt(0.5, color);
-        QBrush gradBrush(grad);
-        painter->fillRect(x2 - iconSize * 2, y1,
-            iconSize * 2, y2 - y1 + 1, gradBrush);
-        QPixmap revertIcon(QIcon::fromTheme(QLatin1String("edit-undo")).pixmap(iconSize, iconSize));
+        if (modified) {
+            alteredOption.rect.setRight( alteredOption.rect.right() - iconSize * 3 / 2 );
+            int y1 = alteredOption.rect.top();
+            QLinearGradient grad(x2 - iconSize * 2, y1, x2 - iconSize / 2, y1);
+            QColor color(
+                alteredOption.palette.color(
+                    (alteredOption.state & QStyle::State_Selected) ? QPalette::Highlight : QPalette::Base ));
+            color.setAlpha(0);
+            grad.setColorAt(0.0, color);
+            color.setAlpha(255);
+            grad.setColorAt(0.5, color);
+            QBrush gradBrush(grad);
+            painter->fillRect(x2 - iconSize * 2, y1, iconSize * 2, y2 - y1 + 1, gradBrush);
 
-        //!TODO
-        //revertIcon = KIconEffect().apply(revertIcon, KIconEffect::Colorize, 1.0,
-        //    alteredOption.palette.color(
-        //        (alteredOption.state & QStyle::State_Selected) ? QPalette::HighlightedText : QPalette::Text ), false);
-        //painter->drawPixmap( x2 - iconSize - 2,
-        //   y1 + 1 + (alteredOption.rect.height() - revertIcon.height()) / 2, revertIcon);
+            //!TODO
+            //QPixmap revertIcon(QIcon::fromTheme(QLatin1String("edit-undo")).pixmap(iconSize, iconSize));
+            //revertIcon = KIconEffect().apply(revertIcon, KIconEffect::Colorize, 1.0,
+            //    alteredOption.palette.color(
+            //        (alteredOption.state & QStyle::State_Selected) ? QPalette::HighlightedText : QPalette::Text ), false);
+            //painter->drawPixmap( x2 - iconSize - 2,
+            //   y1 + 1 + (alteredOption.rect.height() - revertIcon.height()) / 2, revertIcon);
+        }
     }
 
     if (gridLineColor.isValid()) {
         QPen pen(gridLineColor);
         painter->setPen(pen);
-        painter->drawRect(r);
+        painter->drawLine(r.topLeft(), r.topRight() + QPoint(1, 0));
+        painter->drawLine(r.bottomLeft() + QPoint(0, 1), r.bottomRight() + QPoint(1, 1));
+        if (!isGroupHeader) {
+            painter->drawLine(r.topRight() + QPoint(1, 0), r.bottomRight() + QPoint(1, 1));
+            painter->drawLine(r.topLeft(), r.bottomLeft() + QPoint(0, 1));
+        }
     }
     else {
         QPen pen(alteredOption.palette.color(QPalette::AlternateBase));
@@ -206,7 +216,7 @@ QWidget * ItemDelegate::createEditor(QWidget * parent,
     const QStyleOptionViewItem & option, const QModelIndex & index ) const
 {
     if (!index.isValid())
-        return 0;
+        return nullptr;
     const KProperty *property = KPropertyUtils::propertyForIndex(index);
     if (property && property->isReadOnly()) {
         return nullptr;
@@ -230,7 +240,9 @@ QWidget * ItemDelegate::createEditor(QWidget * parent,
     }
     QObject::disconnect(w, SIGNAL(commitData(QWidget*)),
         this, SIGNAL(commitData(QWidget*)));
-    if (property && computeAutoSync( property, static_cast<KPropertyEditorView*>(this->parent())->isAutoSync() )) {
+    if (property && effectiveValueSyncPolicy(property,
+            qobject_cast<KPropertyEditorView*>(this->parent())->isValueSyncEnabled()))
+    {
         QObject::connect(w, SIGNAL(commitData(QWidget*)),
             this, SIGNAL(commitData(QWidget*)));
     }
@@ -240,28 +252,69 @@ QWidget * ItemDelegate::createEditor(QWidget * parent,
 
 //----------
 
-class KPropertyEditorView::Private
+class Q_DECL_HIDDEN KPropertyEditorView::Private
 {
 public:
-    Private()
-     : set(0)
-     , model(0)
+    explicit Private(KPropertyEditorView *view)
+     : model(nullptr)
      , gridLineColor( KPropertyEditorView::defaultGridLineColor() )
-     , autoSync(true)
+     , valueSync(true)
      , slotPropertyChangedEnabled(true)
+     , q(view)
     {
     }
-    KPropertySet *set;
+
+    //! Expands group and parent property items if needed (based on settings)
+    void expandIfNeeded() {
+        if (!model) {
+            return;
+        }
+        const int rowCount = model->rowCount();
+        for (int row = 0; row < rowCount; row++) {
+            expandChildItemsIfNeeded(model->index(row, 0));
+        }
+    }
+
+    //! Expands property child items in a subtree recursively if needed (based on settings)
+    void expandChildItemsIfNeeded(const QModelIndex &parent) {
+        if (!model) {
+            return;
+        }
+        const bool isGroupHeader(model->data(parent, KPropertyEditorDataModel::PropertyGroupRole).toBool());
+        if (isGroupHeader) {
+            if (groupItemsExpanded) {
+                q->expand(parent);
+            }
+        } else {
+            if (childPropertyItemsExpanded) {
+                q->expand(parent);
+            }
+        }
+        const int rowCount = model->rowCount(parent);
+        for (int row = 0; row < rowCount; row++) {
+            const QModelIndex child(model->index(row, 0, parent));
+            expandChildItemsIfNeeded(child);
+        }
+    }
+
+    QPointer<KPropertySet> set;
     KPropertyEditorDataModel *model;
     ItemDelegate *itemDelegate;
     QColor gridLineColor;
-    bool autoSync;
+    bool valueSync;
     bool slotPropertyChangedEnabled;
+    bool childPropertyItemsExpanded = true;
+    bool groupItemsExpanded = true;
+    bool groupsVisible = true;
+    bool toolTipsVisible = false;
+
+private:
+    KPropertyEditorView * const q;
 };
 
 KPropertyEditorView::KPropertyEditorView(QWidget* parent)
         : QTreeView(parent)
-        , d( new Private )
+        , d(new Private(this))
 {
     setObjectName(QLatin1String("KPropertyEditorView"));
     setAlternatingRowColors(true);
@@ -350,7 +403,7 @@ void KPropertyEditorView::changeSetInternal(KPropertySet *set, SetOptions option
 
     QByteArray selectedPropertyName1 = propertyToSelect;
     QByteArray selectedPropertyName2 = propertyToSelect;
-    if (options & PreservePreviousSelection) {
+    if (options & SetOption::PreservePreviousSelection) {
         //try to find prev. selection:
         //1. in new list's prev. selection
         if (set)
@@ -377,16 +430,15 @@ void KPropertyEditorView::changeSetInternal(KPropertySet *set, SetOptions option
 
     KPropertyEditorDataModel *oldModel = d->model;
     const KPropertySetIterator::Order setOrder
-        = (options & AlphabeticalOrder) ? KPropertySetIterator::AlphabeticalOrder : KPropertySetIterator::InsertionOrder;
-    d->model = d->set ? new KPropertyEditorDataModel(d->set, this, setOrder) : 0;
+        = (options & SetOption::AlphabeticalOrder)
+            ? KPropertySetIterator::Order::Alphabetical
+            : KPropertySetIterator::Order::Insertion;
+    d->model = d->set ? new KPropertyEditorDataModel(this, setOrder) : nullptr;
     setModel( d->model );
     delete oldModel;
 
-    if (d->model && d->set && !d->set->isEmpty() && (options & ExpandChildItems)) {
-        const int rowCount = d->model->rowCount();
-        for (int row = 0; row < rowCount; row++) {
-            expand( d->model->index(row, 0) );
-        }
+    if (d->model && d->set && !d->set->isEmpty()) {
+        d->expandIfNeeded();
     }
 
     emit propertySetChanged(d->set);
@@ -408,12 +460,12 @@ void KPropertyEditorView::changeSetInternal(KPropertySet *set, SetOptions option
 
 void KPropertyEditorView::slotSetWillBeCleared()
 {
-    changeSet(0, QByteArray());
+    changeSet(nullptr, QByteArray());
 }
 
 void KPropertyEditorView::slotSetWillBeDeleted()
 {
-    changeSet(0, QByteArray());
+    changeSet(nullptr, QByteArray());
 }
 
 void KPropertyEditorView::slotReadOnlyFlagChanged()
@@ -426,14 +478,51 @@ void KPropertyEditorView::slotReadOnlyFlagChanged()
     }
 }
 
-void KPropertyEditorView::setAutoSync(bool enable)
+void KPropertyEditorView::setValueSyncEnabled(bool set)
 {
-    d->autoSync = enable;
+    d->valueSync = set;
 }
 
-bool KPropertyEditorView::isAutoSync() const
+bool KPropertyEditorView::isValueSyncEnabled() const
 {
-    return d->autoSync;
+    return d->valueSync;
+}
+
+void KPropertyEditorView::setChildPropertyItemsExpanded(bool set)
+{
+    d->childPropertyItemsExpanded = set;
+}
+
+bool KPropertyEditorView::childPropertyItemsExpanded() const
+{
+    return d->childPropertyItemsExpanded;
+}
+
+void KPropertyEditorView::setGroupItemsExpanded(bool set)
+{
+    d->groupItemsExpanded = set;
+}
+
+bool KPropertyEditorView::groupItemsExpanded() const
+{
+    return d->groupItemsExpanded;
+}
+
+bool KPropertyEditorView::groupsVisible() const
+{
+    return d->groupsVisible;
+}
+
+void KPropertyEditorView::setGroupsVisible(bool set)
+{
+    if (d->groupsVisible == set) {
+        return;
+    }
+    if (d->model) {
+        d->model->updateGroupsVisibility();
+        d->expandIfNeeded();
+    }
+    viewport()->update();
 }
 
 void KPropertyEditorView::currentChanged( const QModelIndex & current, const QModelIndex & previous )
@@ -464,9 +553,32 @@ void KPropertyEditorView::drawBranches( QPainter * painter, const QRect & rect,
     QTreeView::drawBranches( painter, rect, index );
 }
 
+void KPropertyEditorView::drawRow(QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index) const
+{
+    if (!d->model) {
+        return;
+    }
+    const KPropertyUtilsPrivate::PainterSaver saver(painter);
+    const bool isGroupHeader(d->model->data(index, KPropertyEditorDataModel::PropertyGroupRole).toBool());
+    QStyleOptionViewItem alteredOption(option);
+    QTreeView::drawRow(painter, alteredOption, index);
+    if (isGroupHeader) {
+        // Special case: group header should be displayed over both columns. There's an issue with
+        // alternate background which is painted over text in the 2nd column, so draw the text here
+        // by hand.
+        QFont font(alteredOption.font);
+        font.setBold(true);
+        alteredOption.font = font;
+        painter->setFont(font);
+        painter->drawText(
+            alteredOption.rect.adjusted(style()->pixelMetric(QStyle::PM_TreeViewIndentation), 0, 0, 0),
+            index.data(Qt::DisplayRole).toString(), Qt::AlignLeft | Qt::AlignVCenter);
+    }
+}
+
 QRect KPropertyEditorView::revertButtonArea( const QModelIndex& index ) const
 {
-    if (index.column() != 0)
+    if (index.column() != 0 || !d->model)
         return QRect();
     QVariant modifiedVariant( d->model->data(index, KPropertyEditorDataModel::PropertyModifiedRole) );
     if (!modifiedVariant.isValid() || !modifiedVariant.toBool())
@@ -500,12 +612,13 @@ void KPropertyEditorView::mousePressEvent ( QMouseEvent * event )
 
 void KPropertyEditorView::undo()
 {
-    if (!d->set || d->set->isReadOnly())
+    if (!d->set || d->set->isReadOnly() || !d->model)
         return;
 
     KProperty *property = d->model->propertyForIndex(currentIndex());
-    if (computeAutoSync( property, d->autoSync ))
+    if (effectiveValueSyncPolicy(property, d->valueSync)) {
         property->resetValue();
+    }
 }
 
 void KPropertyEditorView::acceptInput()
@@ -515,9 +628,7 @@ void KPropertyEditorView::acceptInput()
 
 void KPropertyEditorView::commitData( QWidget * editor )
 {
-    d->slotPropertyChangedEnabled = false;
     QAbstractItemView::commitData( editor );
-    d->slotPropertyChangedEnabled = true;
 }
 
 bool KPropertyEditorView::viewportEvent( QEvent * event )
@@ -543,7 +654,7 @@ QSize KPropertyEditorView::sizeHint() const
 
 KPropertySet* KPropertyEditorView::propertySet() const
 {
-    return d->model ? d->model->propertySet() : nullptr;
+    return d->set;
 }
 
 QColor KPropertyEditorView::gridLineColor() const
@@ -581,7 +692,7 @@ static QModelIndex findChildItem(const KProperty& property, const QModelIndex &p
 void KPropertyEditorView::slotPropertyChanged(KPropertySet& set, KProperty& property)
 {
     Q_UNUSED(set);
-    if (!d->slotPropertyChangedEnabled)
+    if (!d->slotPropertyChangedEnabled || !d->model)
         return;
     d->slotPropertyChangedEnabled = false;
     KProperty *realProperty = &property;
@@ -598,14 +709,14 @@ void KPropertyEditorView::slotPropertyChanged(KPropertySet& set, KProperty& prop
 
 void KPropertyEditorView::updateSubtree(const QModelIndex &index)
 {
-    if (!index.isValid()) {
+    if (!index.isValid() || !d->model) {
         return;
     }
     update(index);
-    QModelIndex valueIndex = d->model->indexForColumn(index, 1);
-    if (valueIndex.isValid()) {
-        update(valueIndex);
-    }
+    update(index.parent());
+    update(d->model->indexForColumn(index, 1));
+    update(d->model->indexForColumn(index.parent(), 1));
+
     KProperty *property = static_cast<KProperty*>(index.internalPointer());
     if (property->children()) {
         int row = 0;
@@ -621,3 +732,14 @@ void KPropertyEditorView::slotPropertyReset(KPropertySet& set, KProperty& proper
 //! @todo OK?
     slotPropertyChanged(set, property);
 }
+
+
+bool KPropertyEditorView::toolTipsVisible() const
+{
+    return d->toolTipsVisible;
+}
+
+void KPropertyEditorView::setToolTipsVisible(bool set)
+{
+    d->toolTipsVisible = set;
+}
